@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2019, 2021, 2022, 2023,
+ * Copyright (C) 1986, 1988, 1989, 1991-2019, 2021-2025,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -414,7 +414,7 @@ opcode2str(OPCODE op)
 	if (op >= Op_illegal && op < Op_final)
 		return optypes[(int) op].desc;
 	fatal(_("unknown opcode %d"), (int) op);
-	return NULL;
+	return "";	// keeps the compiler happy in some cases
 }
 
 /* op2str --- convert an opcode type to corresponding operator or keyword */
@@ -504,12 +504,21 @@ genflags2str(int flagval, const struct flagtab *tab)
 
 /* posix_compare --- compare strings using strcoll */
 
+/*
+ * 1/2026:
+ * On systems except MingGW (even Cygwin), using strcoll() to do the
+ * work looks to be enough.  So that's the default. On MinGW, we do
+ * different tests based on gawk_mb_cur_max.
+ */
+
 static int
 posix_compare(NODE *s1, NODE *s2)
 {
-	int ret;
+	int ret = 0;
 
+#ifdef __MINGW32__
 	if (gawk_mb_cur_max == 1) {
+#endif
 		char save1, save2;
 		const char *p1, *p2;
 
@@ -546,10 +555,10 @@ posix_compare(NODE *s1, NODE *s2)
 
 		s1->stptr[s1->stlen] = save1;
 		s2->stptr[s2->stlen] = save2;
-	}
-	else {
+#ifdef __MINGW32__
+	} else {
 		/* Similar logic, using wide characters */
-		const wchar_t *p1, *p2;
+		const char32_t *p1, *p2;
 
 		(void) force_wstring(s1);
 		(void) force_wstring(s2);
@@ -560,11 +569,11 @@ posix_compare(NODE *s1, NODE *s2)
 		for (;;) {
 			size_t len;
 
-			ret = wcscoll(p1, p2);
+			ret = c32scoll(p1, p2);
 			if (ret != 0)
 				break;
 
-			len = wcslen(p1);
+			len = c32slen(p1);
 			p1 += len + 1;
 			p2 += len + 1;
 
@@ -579,6 +588,7 @@ posix_compare(NODE *s1, NODE *s2)
 			}
 		}
 	}
+#endif /* __MINGW32__ */
 
 	return ret;
 }
@@ -651,10 +661,10 @@ push_frame(NODE *f)
 	fcall_count++;
 	if (fcall_list == NULL) {
 		max_fcall = 10;
-		emalloc(fcall_list, NODE **, (max_fcall + 1) * sizeof(NODE *), "push_frame");
+		emalloc(fcall_list, NODE **, (max_fcall + 1) * sizeof(NODE *));
 	} else if (fcall_count == max_fcall) {
 		max_fcall *= 2;
-		erealloc(fcall_list, NODE **, (max_fcall + 1) * sizeof(NODE *), "push_frame");
+		erealloc(fcall_list, NODE **, (max_fcall + 1) * sizeof(NODE *));
 	}
 
 	if (fcall_count > 1)
@@ -826,9 +836,9 @@ set_OFS()
 	new_ofs_len = OFS_node->var_value->stlen;
 
 	if (OFS == NULL)
-		emalloc(OFS, char *, new_ofs_len + 1, "set_OFS");
+		emalloc(OFS, char *, new_ofs_len + 1);
 	else if (OFSlen < new_ofs_len)
-		erealloc(OFS, char *, new_ofs_len + 1, "set_OFS");
+		erealloc(OFS, char *, new_ofs_len + 1);
 
 	memcpy(OFS, OFS_node->var_value->stptr, OFS_node->var_value->stlen);
 	OFSlen = new_ofs_len;
@@ -900,7 +910,7 @@ fmt_index(NODE *n)
 	char save;
 
 	if (fmt_list == NULL)
-		emalloc(fmt_list, NODE **, fmt_num*sizeof(*fmt_list), "fmt_index");
+		emalloc(fmt_list, NODE **, fmt_num*sizeof(*fmt_list));
 	n = force_string(n);
 
 	save = n->stptr[n->stlen];
@@ -923,7 +933,7 @@ fmt_index(NODE *n)
 
 	if (fmt_hiwater >= fmt_num) {
 		fmt_num *= 2;
-		erealloc(fmt_list, NODE **, fmt_num * sizeof(*fmt_list), "fmt_index");
+		erealloc(fmt_list, NODE **, fmt_num * sizeof(*fmt_list));
 	}
 	fmt_list[fmt_hiwater] = dupnode(n);
 	return fmt_hiwater++;
@@ -1013,6 +1023,9 @@ update_ERRNO_int(int errcode)
 {
 	const char *cp;
 
+	if (do_traditional)
+		return;
+
 	update_PROCINFO_num("errno", errcode);
 	if (errcode) {
 		cp = strerror(errcode);
@@ -1028,6 +1041,9 @@ update_ERRNO_int(int errcode)
 void
 update_ERRNO_string(const char *string)
 {
+	if (do_traditional)
+		return;
+
 	update_PROCINFO_num("errno", 0);
 	unref(ERRNO_node->var_value);
 	size_t len = strlen(string);
@@ -1054,6 +1070,9 @@ update_ERRNO_string(const char *string)
 void
 unset_ERRNO(void)
 {
+	if (do_traditional)
+		return;
+
 	update_PROCINFO_num("errno", 0);
 	unref(ERRNO_node->var_value);
 	ERRNO_node->var_value = dupnode(Nnull_string);
@@ -1129,7 +1148,7 @@ STACK_ITEM *
 grow_stack()
 {
 	STACK_SIZE *= 2;
-	erealloc(stack_bottom, STACK_ITEM *, STACK_SIZE * sizeof(STACK_ITEM), "grow_stack");
+	erealloc(stack_bottom, STACK_ITEM *, STACK_SIZE * sizeof(STACK_ITEM));
 	stack_top = stack_bottom + STACK_SIZE - 1;
 	stack_ptr = stack_bottom + STACK_SIZE / 2;
 	return stack_ptr;
@@ -1283,7 +1302,7 @@ setup_frame(INSTRUCTION *pc)
 	arg_count = (pc + 1)->expr_count;
 
 	if (pcount > 0) {
-		ezalloc(sp, NODE **, pcount * sizeof(NODE *), "setup_frame");
+		ezalloc(sp, NODE **, pcount * sizeof(NODE *));
 	}
 
 	/* check for extra args */
@@ -1317,8 +1336,11 @@ setup_frame(INSTRUCTION *pc)
 
 		/* $0 needs to be passed by value to a function */
 		if (m == fields_arr[0]) {
+			NODE *copy;
+
+			copy = dupnode(m);	// will be real copy
 			DEREF(m);
-			m = dupnode(m);
+			m = copy;
 		}
 
 		switch (m->type) {
@@ -1344,6 +1366,13 @@ setup_frame(INSTRUCTION *pc)
 			r->var_value = dupnode(Nnull_string);
 			break;
 
+		case Node_regex:
+		case Node_dynregex:
+			// 1/2025:
+			// These are weird; they can happen through
+			// indirect calls to some of the builtins, so
+			// handle them if we get them by ....
+			// ... falling through! Yay!
 		case Node_val:
 			r->type = Node_var;
 			r->var_value = m;
@@ -1379,6 +1408,7 @@ setup_frame(INSTRUCTION *pc)
 
 	/* setup new frame */
 	getnode(frame_ptr);
+	memset(frame_ptr, '\0', sizeof(NODE));
 	frame_ptr->type = Node_frame;
 	frame_ptr->stack = sp;
 	frame_ptr->prev_frame_size = (stack_ptr - stack_bottom); /* size of the previous stack frame */
@@ -1542,7 +1572,7 @@ cmp_scalars(scalar_cmp_t comparison_type)
 {
 	NODE *t1, *t2;
 	int di;
-	bool ret;
+	bool ret = false;
 
 	t2 = POP_SCALAR();
 	t1 = TOP();
@@ -1550,12 +1580,15 @@ cmp_scalars(scalar_cmp_t comparison_type)
 	t1 = elem_new_to_scalar(t1);
 	t2 = elem_new_to_scalar(t2);
 
+	t1 = fixtype(t1);
+	t2 = fixtype(t2);
+
 	if (t1->type == Node_var_array) {
 		DEREF(t2);
 		fatal(_("attempt to use array `%s' in a scalar context"), array_vname(t1));
 	}
 
-	if ((t1->flags & STRING) != 0 || (t2->flags & STRING) != 0) {
+	if ((t1->flags & (STRING|REGEX)) != 0 || (t2->flags & (STRING|REGEX)) != 0) {
 		bool use_strcmp = (comparison_type == SCALAR_EQ || comparison_type == SCALAR_NEQ);
 		di = cmp_nodes(t1, t2, use_strcmp);
 
@@ -1578,11 +1611,11 @@ cmp_scalars(scalar_cmp_t comparison_type)
 		case SCALAR_GE:
 			ret = (di >= 0);
 			break;
+		default:
+			cant_happen("invalid value %d in cmp_scalars", (int) comparison_type);
+			break;
 		}
 	} else {
-		fixtype(t1);
-		fixtype(t2);
-
 #ifdef HAVE_MPFR
 		if (do_mpfr)
 			ret = mpg_cmp_as_numbers(t1, t2, comparison_type);
@@ -1609,7 +1642,7 @@ cmp_doubles(const NODE *t1, const NODE *t2, scalar_cmp_t comparison_type)
 
 	bool t1_nan = isnan(t1->numbr);
 	bool t2_nan = isnan(t2->numbr);
-	int ret;
+	int ret = false;
 
 	if ((t1_nan || t2_nan) && comparison_type != SCALAR_NEQ)
 		return false;
@@ -1633,6 +1666,9 @@ cmp_doubles(const NODE *t1, const NODE *t2, scalar_cmp_t comparison_type)
 	case SCALAR_GE:
 		ret = (t1->numbr >= t2->numbr);
 		break;
+	default:
+		cant_happen("invalid value %d in cmp_doubles", (int) comparison_type);
+		break;
 	}
 
 	return ret;
@@ -1711,6 +1747,7 @@ PUSH_CODE(INSTRUCTION *cp)
 {
 	NODE *r;
 	getnode(r);
+	memset(r, '\0', sizeof(NODE));
 	r->type = Node_instruction;
 	r->code_ptr = cp;
 	PUSH(r);
@@ -1766,7 +1803,7 @@ push_exec_state(INSTRUCTION *cp, int rule, char *src, STACK_ITEM *sp)
 {
 	EXEC_STATE *es;
 
-	emalloc(es, EXEC_STATE *, sizeof(EXEC_STATE), "push_exec_state");
+	emalloc(es, EXEC_STATE *, sizeof(EXEC_STATE));
 	es->rule = rule;
 	es->cptr = cp;
 	es->stack_size = (sp - stack_bottom) + 1;
@@ -1857,12 +1894,13 @@ init_interpret()
 	if ((newval = getenv_long("GAWK_STACKSIZE")) > 0)
 		STACK_SIZE = newval;
 
-	emalloc(stack_bottom, STACK_ITEM *, STACK_SIZE * sizeof(STACK_ITEM), "grow_stack");
+	emalloc(stack_bottom, STACK_ITEM *, STACK_SIZE * sizeof(STACK_ITEM));
 	stack_ptr = stack_bottom - 1;
 	stack_top = stack_bottom + STACK_SIZE - 1;
 
 	/* initialize frame pointer */
 	getnode(frame_ptr);
+	memset(frame_ptr, '\0', sizeof(NODE));
 	frame_ptr->type = Node_frame;
 	frame_ptr->stack = NULL;
 	frame_ptr->func_node = NULL;	/* in main */
@@ -1888,6 +1926,21 @@ init_interpret()
 		interpret = r_interpret;
 }
 
+/* elem_new_reset --- clear the elemnew_parent and elemnew_vname fields of a Node_elem_new. */
+
+void
+elem_new_reset(NODE *n)
+{
+	assert(n->type == Node_elem_new);
+
+	if (n->elemnew_vname != NULL) {
+		efree(n->elemnew_vname);
+		n->elemnew_vname = NULL;
+	}
+	n->elemnew_parent = NULL;
+	n->vname = NULL;
+}
+
 /* elem_new_to_scalar --- convert Node_elem_new to untyped scalar */
 
 NODE *
@@ -1896,6 +1949,8 @@ elem_new_to_scalar(NODE *n)
 	if (n->type != Node_elem_new)
 		return n;
 
+	elem_new_reset(n);
+
 	if (n->valref > 1) {
 		unref(n);
 		return dupnode(Nnull_string);
