@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 2012, 2013, 2015, 2017, 2018, 2019, 2021
+ * Copyright (C) 2012, 2013, 2015, 2017, 2018, 2019, 2021, 2022, 2024, 2025,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -265,7 +265,7 @@ mpg_zero(NODE *n)
 /* force_mpnum --- force a value to be a GMP integer or MPFR float */
 
 static int
-force_mpnum(NODE *n, int do_nondec, int use_locale)
+force_mpnum(NODE *n, bool do_nondec, bool use_locale)
 {
 	char *cp, *cpend, *ptr, *cp1;
 	char save;
@@ -307,6 +307,9 @@ force_mpnum(NODE *n, int do_nondec, int use_locale)
 	if (do_nondec)
 		base = get_numbase(cp1, cpend - cp1, use_locale);
 
+	if (base == 16 && (strchr(cp, 'p') != NULL || strchr(cp, 'P') != NULL))
+		goto isfloat;
+
 	if (base != 10 || ! mpg_maybe_float(cp1, use_locale)) {
 		mpg_zero(n);
 		errno = 0;
@@ -316,6 +319,7 @@ force_mpnum(NODE *n, int do_nondec, int use_locale)
 		goto done;
 	}
 
+isfloat:
 	if (is_mpg_integer(n)) {
 		mpz_clear(n->mpg_i);
 		n->flags &= ~MPZN;
@@ -349,8 +353,21 @@ mpg_force_number(NODE *n)
 {
 	char *cp, *cpend;
 
+	if (n->type == Node_elem_new) {
+		elem_new_reset(n);
+		n->type = Node_val;
+
+		return n;
+	}
+
 	if ((n->flags & NUMCUR) != 0)
 		return n;
+
+	/*
+	 * We should always set NUMCUR. If USER_INPUT is set and it's a
+	 * numeric string, we clear STRING and enable NUMBER, but if it's not
+	 * numeric, we disable USER_INPUT.
+	 */
 	n->flags |= NUMCUR;
 
 	/* Trim leading white space, bailing out if there's nothing else */
@@ -388,7 +405,7 @@ mpg_force_number(NODE *n)
 	/* else POSIX, so
 		fall through */
 
-	if (force_mpnum(n, (do_non_decimal_data && ! do_traditional), true)) {
+	if (force_mpnum(n, (do_non_decimal_data || do_posix), true)) {
 		if ((n->flags & USER_INPUT) != 0) {
 			/* leave USER_INPUT set to indicate a strnum */
 			n->flags &= ~STRING;
@@ -416,16 +433,16 @@ mpg_format_val(const char *format, int index, NODE *s)
 		return make_string(result, strlen(result));
 	}
 
-	/* create dummy node for a sole use of format_tree */
+	/* create dummy node for a sole use of format_args */
 	dummy[1] = s;
 	oflags = s->flags;
 
 	if (is_mpg_integer(s) || mpfr_integer_p(s->mpg_numbr)) {
 		/* integral value, use %d */
-		r = format_tree("%d", 2, dummy, 2);
+		r = format_args("%d", 2, dummy, 2);
 		s->stfmt = STFMT_UNUSED;
 	} else {
-		r = format_tree(format, fmt_list[index]->stlen, dummy, 2);
+		r = format_args(format, fmt_list[index]->stlen, dummy, 2);
 		assert(r != NULL);
 		s->stfmt = index;
 	}
@@ -476,6 +493,53 @@ mpg_cmp(const NODE *t1, const NODE *t2)
 	return cmp_awknums(t1, t2);
 }
 
+/* mpg_cmp_as_numbers --- compare two numbers, similar to doubles */
+
+bool
+mpg_cmp_as_numbers(const NODE *t1, const NODE *t2, scalar_cmp_t comparison_type)
+{
+	/*
+	 * This routine provides numeric comparisons that should work
+	 * the same as in C.  It should NOT be used for sorting.
+	 */
+
+	bool t1_nan = mpfr_nan_p(t1->mpg_numbr);
+	bool t2_nan = mpfr_nan_p(t2->mpg_numbr);
+	bool ret = false;
+
+	// MPFR is different than native doubles...
+	if (t1_nan || t2_nan)
+		return comparison_type == SCALAR_NEQ;
+
+	int di = mpg_cmp(t1, t2);
+
+	switch (comparison_type) {
+	case SCALAR_EQ:
+		ret = (di == 0);
+		break;
+	case SCALAR_NEQ:
+		ret = (di != 0);
+		break;
+	case SCALAR_LT:
+		ret = (di < 0);
+		break;
+	case SCALAR_LE:
+		ret = (di <= 0);
+		break;
+	case SCALAR_GT:
+		ret = (di > 0);
+		break;
+	case SCALAR_GE:
+		ret = (di >= 0);
+		break;
+	default:
+		cant_happen("invalid comparison type %d", (int) comparison_type);
+		break;
+	}
+
+	return ret;
+}
+
 
 /*
  * mpg_update_var --- update NR or FNR.
@@ -496,7 +560,7 @@ mpg_update_var(NODE *n)
 		nr = FNR;
 		nq = MFNR;
 	} else
-		cant_happen();
+		cant_happen("invalid node for mpg_update_var%s", "");
 
 	if (mpz_sgn(nq) == 0) {
 		/* Efficiency hack similar to that for AWKNUM */
@@ -528,7 +592,7 @@ mpg_set_var(NODE *n)
 	else if (n == FNR_node)
 		nq = MFNR;
 	else
-		cant_happen();
+		cant_happen("invalid node for mpg_set_var%s", "");
 
 	if (is_mpg_integer(val))
 		r = val->mpg_i;
@@ -720,6 +784,8 @@ do_mpfr_atan2(int nargs)
 	mpfr_ptr p1, p2;
 	int tval;
 
+	check_exact_args(nargs, "atan2", 2);
+
 	t2 = POP_SCALAR();
 	t1 = POP_SCALAR();
 
@@ -756,6 +822,8 @@ do_mpfr_func(const char *name,
 	int tval;
 	mpfr_prec_t argprec;
 
+	check_exact_args(nargs, name, 1);
+
 	t1 = POP_SCALAR();
 	if (do_lint && (fixtype(t1)->flags & NUMBER) == 0)
 		lintwarn(_("%s: received non-numeric argument"), name);
@@ -827,6 +895,8 @@ do_mpfr_int(int nargs)
 {
 	NODE *tmp, *r;
 
+	check_exact_args(nargs, "int", 1);
+
 	tmp = POP_SCALAR();
 	if (do_lint && (fixtype(tmp)->flags & NUMBER) == 0)
 		lintwarn(_("int: received non-numeric argument"));
@@ -857,6 +927,8 @@ do_mpfr_compl(int nargs)
 	NODE *tmp, *r;
 	mpz_ptr zptr;
 
+	check_exact_args(nargs, "compl", 1);
+
 	tmp = POP_SCALAR();
 	if (do_lint && (fixtype(tmp)->flags & NUMBER) == 0)
 		lintwarn(_("compl: received non-numeric argument"));
@@ -919,7 +991,7 @@ get_intval(NODE *t1, int argnum, const char *op)
                                 	op, argnum, left)
 				);
 
-			emalloc(pz, mpz_ptr, sizeof (mpz_t), "get_intval");
+			emalloc(pz, mpz_ptr, sizeof (mpz_t));
 			mpz_init(pz);
 			return pz;	/* should be freed */
 		}
@@ -938,7 +1010,7 @@ get_intval(NODE *t1, int argnum, const char *op)
 				);
 		}
 
-		emalloc(pz, mpz_ptr, sizeof (mpz_t), "get_intval");
+		emalloc(pz, mpz_ptr, sizeof (mpz_t));
 		mpz_init(pz);
 		mpfr_get_z(pz, left, MPFR_RNDZ);	/* float to integer conversion */
 		return pz;	/* should be freed */
@@ -976,6 +1048,8 @@ do_mpfr_lshift(int nargs)
 	unsigned long shift;
 	mpz_ptr pz1, pz2;
 
+	check_exact_args(nargs, "lshift", 2);
+
 	t2 = POP_SCALAR();
 	t1 = POP_SCALAR();
 
@@ -1008,6 +1082,8 @@ do_mpfr_rshift(int nargs)
 	unsigned long shift;
 	mpz_ptr pz1, pz2;
 
+	check_exact_args(nargs, "rshift", 2);
+
 	t2 = POP_SCALAR();
 	t1 = POP_SCALAR();
 
@@ -1128,6 +1204,8 @@ do_mpfr_strtonum(int nargs)
 {
 	NODE *tmp, *r;
 
+	check_exact_args(nargs, "strtonum", 1);
+
 	tmp = fixtype(POP_SCALAR());
 	if ((tmp->flags & NUMBER) == 0) {
 		r = mpg_integer();	/* will be changed to MPFR float if necessary in force_mpnum() */
@@ -1165,6 +1243,8 @@ do_mpfr_rand(int nargs ATTRIBUTE_UNUSED)
 	NODE *res;
 	int tval;
 
+	check_exact_args(nargs, "rand", 0);
+
 	if (firstrand) {
 #if 0
 		/* Choose the default algorithm */
@@ -1215,6 +1295,8 @@ do_mpfr_srand(int nargs)
 		firstrand = false;
 	}
 
+	check_args_min_max(nargs, "srand", 0, 1);
+
 	res = mpg_integer();
 	mpz_set(res->mpg_i, seed);	/* previous seed */
 
@@ -1243,9 +1325,9 @@ do_mpfr_srand(int nargs)
 /*
  * We define the semantics as:
  * 	numerator = int(numerator)
- *	denominator = int(denonmator)
- *	quotient = int(numerator / denomator)
- *	remainder = int(numerator % denomator)
+ *	denominator = int(denominator)
+ *	quotient = int(numerator / denominator)
+ *	remainder = int(numerator % denominator)
  */
 
 NODE *
@@ -1256,6 +1338,8 @@ do_mpfr_intdiv(int nargs)
 	NODE *quotient, *remainder;
 	NODE *sub, **lhs;
 
+	check_exact_args(nargs, "intdiv", 3);
+
 	result = POP_PARAM();
 	if (result->type != Node_var_array)
 		fatal(_("intdiv: third argument is not an array"));
@@ -1518,7 +1602,10 @@ mpg_div(NODE *t1, NODE *t2)
 	return r;
 }
 
-/* mpg_mod --- modulus operation with arbitrary-precision numbers */
+/*
+ * mpg_mod --- modulus operation with arbitrary-precision numbers.
+ *	See test/mpfrrem.awk.
+ */
 
 static NODE *
 mpg_mod(NODE *t1, NODE *t2)
@@ -1527,29 +1614,18 @@ mpg_mod(NODE *t1, NODE *t2)
 	int tval;
 
 	if (is_mpg_integer(t1) && is_mpg_integer(t2)) {
-		/*
-		 * 8/2014: Originally, this was just
-		 *
-		 * r = mpg_integer();
-		 * mpz_mod(r->mpg_i, t1->mpg_i, t2->mpg_i);
-		 *
-		 * But that gave very strange results with negative numerator:
-		 *
-		 *	$ ./gawk -M 'BEGIN { print -15 % 7 }'
-		 *	6
-		 *
-		 * So instead we use mpz_tdiv_qr() to get the correct result
-		 * and just throw away the quotient. We could not find any
-		 * reason why mpz_mod() wasn't working correctly.
-		 */
-		NODE *dummy_quotient;
-
 		if (mpz_sgn(t2->mpg_i) == 0)
 			fatal(_("division by zero attempted"));
 		r = mpg_integer();
-		dummy_quotient = mpg_integer();
-		mpz_tdiv_qr(dummy_quotient->mpg_i, r->mpg_i, t1->mpg_i, t2->mpg_i);
-		unref(dummy_quotient);
+		/*
+		 * Before 2014-08, this called mpz_mod(), which is inconsistent
+		 * with mpfr_fmod() when the numerator is negative. It gave:
+		 *   $ gawk -M 'BEGIN { print 15 % 7, 15 % -7, -15 % 7, -15 % -7 }'
+		 *   1 1 6 6
+		 *   $ gawk -M 'BEGIN { print 15.0 % 7, 15.0 % -7, -15.0 % 7, -15.0 % -7 }'
+		 *   1 1 -1 -1
+		 */
+		mpz_tdiv_r(r->mpg_i, t1->mpg_i, t2->mpg_i);
 	} else {
 		mpfr_ptr p1, p2;
 		p1 = MP_FLOAT(t1);
@@ -1775,8 +1851,17 @@ mod:
 			tval = mpfr_neg(r->mpg_numbr, t1->mpg_numbr, ROUND_MODE);
 			IEEE_FMT(r->mpg_numbr, tval);
 		} else {
-			r = mpg_integer();
-			mpz_neg(r->mpg_i, t1->mpg_i);
+			if (! is_zero(t1)) {
+				r = mpg_integer();
+				mpz_neg(r->mpg_i, t1->mpg_i);
+			} else {
+				// have to convert to MPFR for -0.0. sigh
+				r = mpg_float();
+				tval = mpfr_set_d(r->mpg_numbr, 0.0, ROUND_MODE);
+				IEEE_FMT(r->mpg_numbr, tval);
+				tval = mpfr_neg(r->mpg_numbr, r->mpg_numbr, ROUND_MODE);
+				IEEE_FMT(r->mpg_numbr, tval);
+			}
 		}
 		DEREF(t1);
 		REPLACE(r);
@@ -1827,7 +1912,7 @@ mod:
 			r = mpg_pow(t1, t2);
 			break;
 		default:
-			cant_happen();
+			cant_happen("unexpected opcode %s", opcode2str(op));
 		}
 
 		DEREF(t2);
@@ -1878,6 +1963,38 @@ mpfr_unset(NODE *n)
 		mpz_clear(n->mpg_i);
 }
 
+/*
+ * Custom memory allocation functions for GMP / MPFR. We need these so that the
+ * persistent memory feature will also work with the -M option.
+ *
+ * These just call malloc/realloc/free; if we are using PMA then those are
+ * redefined as macros to point at the pma functions, so all should "just work."
+ */
+
+/* mpfr_mem_alloc --- allocate memory */
+
+void *
+mpfr_mem_alloc(size_t alloc_size)
+{
+	return malloc(alloc_size);
+}
+
+/* mpfr_mem_realloc --- reallocate memory */
+
+void *
+mpfr_mem_realloc(void *ptr, size_t old_size, size_t new_size)
+{
+	return realloc(ptr, new_size);
+}
+
+/* mpfr_mem_free --- free memory */
+
+void
+mpfr_mem_free(void *ptr, size_t size)
+{
+	free(ptr);
+}
+
 #else
 
 void
