@@ -29,35 +29,10 @@
  */
 #include "awk.h"	/* includes config.h for us */
 
-
-#ifndef HAVE_SYSTEM
-#include "missing_d/system.c"
-#endif /* HAVE_SYSTEM */
-
-#ifndef HAVE_MEMCMP
-#include "missing_d/memcmp.c"
-#endif	/* HAVE_MEMCMP */
-
-#ifndef HAVE_MEMCPY
-#include "missing_d/memcpy.c"
-#endif	/* HAVE_MEMCPY */
-
-#ifndef HAVE_MEMSET
-#include "missing_d/memset.c"
-#endif	/* HAVE_MEMSET */
-
-#ifndef HAVE_MEMMOVE
-#include "missing_d/memmove.c"
-#endif	/* HAVE_MEMMOVE */
-
 #if !defined(HAVE_STRNCASECMP) || !defined(HAVE_STRCASECMP)
 #include "missing_d/strncasecmp.c"
 #endif	/* HAVE_STRCASE */
 
-#ifndef HAVE_STRERROR
-#include "missing_d/strerror.c"
-#endif	/* HAVE_STRERROR */
-
 #ifndef HAVE_STRFTIME
 # ifdef __MINGW32__
 /* Need to use underlying_strftime in replacement strftime.  */
@@ -67,50 +42,16 @@
 # ifdef __MINGW32__
 #  undef HAVE_STRFTIME
 # endif
-#endif	/* HAVE_STRFTIME */
-
-#ifndef HAVE_STRCHR
-#include "missing_d/strchr.c"
-#endif	/* HAVE_STRCHR */
-
-#if !defined(HAVE_STRTOD)
-#include "missing_d/strtod.c"
-#endif	/* HAVE_STRTOD */
-
-#ifndef HAVE_STRTOUL
-#include "missing_d/strtoul.c"
-#endif	/* HAVE_STRTOUL */
-
-#ifndef HAVE_TZSET
-#include "missing_d/tzset.c"
-#endif /* HAVE_TZSET */
-
-#ifndef HAVE_MKTIME
-/* mktime.c defines main() if DEBUG is set */
-#undef DEBUG
-#include "missing_d/mktime.c"
-#endif /* HAVE_MKTIME */
+#endif  /* HAVE_STRFTIME */
 
 #ifndef HAVE_TIMEGM
 #include "missing_d/timegm.c"
 #endif /* HAVE_TIMEGM */
 
-#ifndef HAVE_SNPRINTF
-#include "missing_d/snprintf.c"
-#endif
-
 #if defined(HAVE_SOCKETS) && ! defined(HAVE_GETADDRINFO)
 #include "missing_d/getaddrinfo.c"
 #endif
 
-#ifndef HAVE_USLEEP
-#include "missing_d/usleep.c"
-#endif
-
-#ifndef HAVE_SETENV
-#include "missing_d/setenv.c"
-#endif
-
 #ifndef HAVE_STRCOLL
 #include "missing_d/strcoll.c"
 #endif
@@ -118,3 +59,13 @@
 #ifndef HAVE_STRSIGNAL
 #include "missing_d/strsignal.c"
 #endif
+
+#if defined(__MINGW32__) && !defined(_UCRT)
+/* We need to work around the MSVCRT bug when writing multibyte
+   strings to the console.  */
+# if !defined(__MINGW32_MAJOR_VERSION) || __MINGW32_MAJOR_VERSION < 5
+   /* MinGW64 has vasprintf, mingw.org's MinGW doesn't. */
+#  define HAVE_VASPRINTF 1
+# endif
+#include "missing_d/stdio-consolesafe.c"
+#endif
