@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2015, 2017-2020, 2022,
+ * Copyright (C) 1986, 1988, 1989, 1991-2015, 2017-2020, 2022, 2023, 2025,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -31,9 +31,6 @@ extern INSTRUCTION *rule_list;
 
 #define HASHSIZE	1021
 
-static int func_count;	/* total number of functions */
-static int var_count;	/* total number of global variables and functions */
-
 static NODE *symbol_list;
 static void (*install_func)(NODE *) = NULL;
 static NODE *make_symbol(const char *name, NODETYPE type);
@@ -50,6 +47,19 @@ NODE *symbol_table, *func_table;
 /* Use a flag to avoid a strcmp() call inside install() */
 static bool installing_specials = false;
 
+// Using persistent memory, manage the root pointer
+// which holds this struct:
+struct root_pointers {
+	NODE *global_table;
+	NODE *func_table;
+	NODE *symbol_table;
+	struct block_header nextfree[BLOCK_MAX];
+	int mpfr;
+	bool first;
+	struct extension *ext_list;
+	const char *version_string;
+} *root_pointers = NULL;
+
 /* init_the_tables --- deal with the tables for in memory use */
 
 static void
@@ -80,13 +90,6 @@ init_symbol_table()
 		return;
 	}
 
-	// using persistent memory, get the root pointer
-	// which holds this struct:
-	struct root_pointers {
-		NODE *global_table;
-		NODE *func_table;
-	} *root_pointers = NULL;
-
 	root_pointers = (struct root_pointers *) pma_get_root();
 
 	if (root_pointers == NULL) {
@@ -94,28 +97,70 @@ init_symbol_table()
 
 		// set up the tables
 		init_the_tables();
+		init_extension_list();
 
 		// save the pointers for the next time.
-		emalloc(root_pointers, struct root_pointers *, sizeof(struct root_pointers), "init_symbol_table");
+		emalloc(root_pointers, struct root_pointers *, sizeof(struct root_pointers));
+		memset(root_pointers, 0, sizeof(struct root_pointers));
 		root_pointers->global_table = global_table;
 		root_pointers->func_table = func_table;
+		root_pointers->symbol_table = symbol_table;
+		root_pointers->ext_list = extension_list;
+		root_pointers->version_string = estrdup(version_string, strlen(version_string));
+		root_pointers->first = true;
+		root_pointers->mpfr = 0;
 		pma_set_root(root_pointers);
 	} else {
 		// this is the next time, get the saved pointers and put them back in place
 		global_table = root_pointers->global_table;
 		func_table = root_pointers->func_table;
+		symbol_table = root_pointers->symbol_table;
+		extension_list = root_pointers->ext_list;
+		memcpy(nextfree, root_pointers->nextfree, sizeof(nextfree));
 
 		// still need to set this one up as usual
 		getnode(param_table);
 		memset(param_table, '\0', sizeof(NODE));
 		null_array(param_table);
-
-		// set the global variables
-		symbol_table = lookup("SYMTAB");
-		func_table = lookup("FUNCTAB");
+		if (strcmp(root_pointers->version_string, version_string) != 0)
+			warning(_("%s was created by version `%s', but the current version is `%s'"),
+					persist_file,
+					root_pointers->version_string,
+					version_string);
 	}
 }
 
+/* pma_mpfr_check --- check that -M is same between invocations */
+
+void
+pma_mpfr_check(void)
+{
+	if (! using_persistent_malloc)
+		return;
+
+	if (root_pointers->first) {
+		root_pointers->first = false;
+		root_pointers->mpfr = do_mpfr;
+		return;
+	}
+
+	if (root_pointers->mpfr != do_mpfr)
+		fatal(_("current setting of -M/--bignum does not match saved setting in PMA backing file"));
+}
+
+/* pma_save_free_lists --- save the free lists in the root pointer */
+
+void
+pma_save_free_lists(void)
+{
+	if (! using_persistent_malloc)
+		return;
+
+	assert(! root_pointers->first);
+
+	memcpy(root_pointers->nextfree, nextfree, sizeof(nextfree));
+}
+
 /*
  * install_symbol:
  * Install a global name in the symbol table, even if it is already there.
@@ -156,7 +201,7 @@ lookup(const char *name)
 		if (assoc_empty(tables[i]))
 			continue;
 
-		if ((do_posix || do_traditional) && tables[i] == global_table)
+		if (do_traditional && tables[i] == global_table)
 			continue;
 
 		n = in_array(tables[i], tmp);
@@ -181,7 +226,7 @@ make_params(char **pnames, int pcount)
 	if (pcount <= 0 || pnames == NULL)
 		return NULL;
 
-	ezalloc(parms, NODE *, pcount * sizeof(NODE), "make_params");
+	ezalloc(parms, NODE *, pcount * sizeof(NODE));
 
 	for (i = 0, p = parms; i < pcount; i++, p++) {
 		p->type = Node_param_list;
@@ -370,10 +415,6 @@ install(const char *name, NODE *parm, NODETYPE type)
 	else {
 		/* global symbol */
 		r = make_symbol(name, type);
-		if (type == Node_func)
-			func_count++;
-		if (type != Node_ext_func && type != Node_builtin_func && table != global_table)
-			var_count++;	/* total, includes Node_func */
 	}
 
 	if (type == Node_param_list) {
@@ -450,7 +491,7 @@ get_symbols(SYMBOL_TYPE what, bool sort)
 		max = the_table->table_size * 2;
 
 		list = assoc_list(the_table, "@unsorted", ASORTI);
-		emalloc(table, NODE **, (func_count + 1) * sizeof(NODE *), "get_symbols");
+		emalloc(table, NODE **, (the_table->table_size + 1) * sizeof(NODE *));
 
 		for (i = count = 0; i < max; i += 2) {
 			r = list[i+1];
@@ -467,7 +508,7 @@ get_symbols(SYMBOL_TYPE what, bool sort)
 
 		list = assoc_list(the_table, "@unsorted", ASORTI);
 		/* add three: one for FUNCTAB, one for SYMTAB, and one for a final NULL */
-		emalloc(table, NODE **, (var_count + 1 + 1 + 1) * sizeof(NODE *), "get_symbols");
+		emalloc(table, NODE **, (the_table->table_size + 1 + 1 + 1) * sizeof(NODE *));
 
 		for (i = count = 0; i < max; i += 2) {
 			r = list[i+1];
@@ -570,6 +611,7 @@ append_symbol(NODE *r)
 	NODE *p;
 
 	getnode(p);
+	memset(p, '\0', sizeof(NODE));
 	p->lnode = r;
 	p->rnode = symbol_list->rnode;
 	symbol_list->rnode = p;
@@ -804,7 +846,7 @@ bcalloc(OPCODE op, int size, int srcline)
 		pool->free_space += size;
 	} else {
 		struct instruction_block *block;
-		emalloc(block, struct instruction_block *, sizeof(struct instruction_block), "bcalloc");
+		emalloc(block, struct instruction_block *, sizeof(struct instruction_block));
 		block->next = pool->block_list;
 		pool->block_list = block;
 		cp = &block->i[0];
@@ -825,7 +867,7 @@ new_context()
 {
 	AWK_CONTEXT *ctxt;
 
-	ezalloc(ctxt, AWK_CONTEXT *, sizeof(AWK_CONTEXT), "new_context");
+	ezalloc(ctxt, AWK_CONTEXT *, sizeof(AWK_CONTEXT));
 	ctxt->srcfiles.next = ctxt->srcfiles.prev = & ctxt->srcfiles;
 	ctxt->rule_list.opcode = Op_list;
 	ctxt->rule_list.lasti = & ctxt->rule_list;
